///////////////////////////////////////////////////////////////////////////////
// 
// CODAGES BINAIRES DES NOMBRES 
// 
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// test si la chaine de caracteres s est formee uniquement de "0" et/ou "1"
// Entree : s = chaine de caracteres
// Sortie : r = booleen
function r = is_string_01(s)
    r = %t
    k = 1
    while k<=length(s) & r
        c = part(s,k);
        r = (c=="0")|(c=="1") 
        k=k+1
    end
endfunction

///////////////////////////////////////////////////////////////////////////////
// codage du nombre entier positif n sur 8 chiffres binaires
// Entrée : n = entier entre 0 et 255
// Sortie : s8 = chaine de 8 caracteres '0' et/ou '1' 
function s8 = codage_entier8(n)
    // test des parametres en entree
    if type(n)~=1 then
        error("n doit etre une valeur numerique")
    end
    if length(n)~=1 then
        error("n doit etre un scalaire et non pas un tableau")
    end
    
    // convertir n en entier et tester la valeur
    n = floor(n);
    if n<0 | n>255 then
        error(sprintf("n doit etre entre 0 et %1.0f", 2^8-1))
    end
    
    // construire la chaine de caracteres s8
    s8="";
    for i=1:8
        r = modulo(n,2);
        n = floor(n/2);
        if r==0 then
            s8 = "0"+s8;
        else
            s8 = "1"+s8;
        end
    end
endfunction

///////////////////////////////////////////////////////////////////////////////
// écriture du nombre entier positif n sur p chiffres binaires
// Entree : n = nombre entier entre 0 et 2^p-1
//          p = nombre entier entre 1 et 52
// Sortie : sp = chaine de p caracteres '0' et/ou '1' 
function sp = codage_entier(n,p)
    // test des parametres en entree
    if type(n)~=1 then
        error("n doit etre une valeur numerique")
    end
    if length(n)~=1 then
        error("n doit etre un scalaire et non pas un tableau")
    end
    
    if type(p)~=1 then
        error("p doit etre une valeur numerique")
    end
    if length(p)~=1 then
        error("p doit etre un scalaire et non pas un tableau")
    end
    
    // convertir p en entier et tester la valeur
    p = floor(p);
    if p<1 | p>52 then
        error("p doit entre 1 et 52")
    end
    
    // convertir n en entier et tester la valeur
    n = floor(n)
    if n<0 | n>2^p-1 then
        error(sprintf("n doit entre 0 et %1.0f", 2^p-1))
    end
    
    // construire la chaine de caracteres sp
    sp="";
    for i=1:p
        r = modulo(n,2);
        n = floor(n/2);
        if r==0 then
            sp = "0"+sp;
        else
            sp = "1"+sp;
        end
    end
    
endfunction

///////////////////////////////////////////////////////////////////////////////
// valeur du nombre dont l'écriture binaire est s
// Entree : s = chaine de caracteres '0' et/ou '1'
// Sortie : v = valeur entiere correspondante
function v = valeur_entier(s)
    // test des parametres en entree
    if type(s)~=10 then
        error("s doit etre une chaine de caracteres")
    end
    if prod(size(s))~=1 then
        error("s doit etre une chaine de caracteres unique")
    end
    if ~is_string_01(s) then
        error("s doit etre composee uniquement de ""0"" ou de ""1""")
    end
    
    // calcul de v
    v = 0;
    for k=1:length(s)
        c = part(s,k)
        if c=="0" then
            v=v*2;
        else
            v=v*2+1;
        end
    end
endfunction

function test_vb2()
    v=valeur_entier("0000000000000000000000000000000000000000000000000000");
    mprintf("v = %20.0f\n", v)
    v=valeur_entier("1111111111111111111111111111111111111111111111111111");
    mprintf("v = %20.0f\n", v)
    v=valeur_entier("11111111111111111111111111111111111111111111111111110");
    mprintf("v = %20.0f\n", v)
    v=valeur_entier("11111111111111111111111111111111111111111111111111111");
    mprintf("v = %20.0f\n", v)
endfunction

/////////////////////////////////////////////////////////////////////////////
// codage du réel flottant v au format binaire 64 chiffres binaires
// Entree : v = réel, ou %inf ou -%inf ou %nan
// Sortie : s = chaine de 1 caracteres  '0' et/ou '1': le signe
//          e = chaine de 11 caracteres '0' et/ou '1': l'exposant
//          m = chaine de 52 caracteres '0' et/ou '1': la mantisse
function [s,e,m] = codage_reel64(v)
    // test des parametres en entree
    if type(v)~=1 then
        error("v doit etre une valeur numerique")
    end
    if length(v)~=1 then
        error("v doit etre un scalaire et non pas un tableau")
    end
    
    // cas Not A Number (%nan)
    if isnan(v) then
        s = '0';
        e = '11111111111';
        m = '1000000000000000000000000000000000000000000000000000';
        return
    end
    
    // cas infini (%inf ou -%inf)
    if isinf(v) then
        if v<0 then
            s = '1';
        else
            s = '0';
        end
        e = '11111111111';
        m = '0000000000000000000000000000000000000000000000000000';
        return
    end
    
    //-- v est une valeur reelle --//
    
    // le signe
    if v<0 then
        s = '1';
        v = -v; // remplacer v par sa valeur absolue
    else
        s = '0';
    end
    
    // cas v < 2^(-1022)
    if v<2^(-1022) then
        e = '00000000000';
        
        // multiplier v par 2^(1074)
        v = v*2^(1000);
        v = v*2^(74);
        m = codage_entier(v,52)
        return;
    end
    
    // cas v >= 2^(-1022)
    // recherche de l'entier p tel que 1 <= v*2^(p) < 2
    // et remplacer v par v/2^(p)
    p=0;
    while v>=2
        p=p+1;
        v=v/2;
    end
    while v<1
        p=p-1;
        v=v*2;
    end
    e = codage_entier(p+1023,11)
    m = codage_entier((v-1)*2^(52),52)
        
endfunction 

//////////////////////////////////////////////////////////////////////////////
// détermination de la valeur dont le codage "reel flottant 64 bits" est
// [s,e,m]
// Entree : s = chaine de 1 caractere
//          e = chaine de 11 caracteres
//          m = chaine de 52 caracteres
//           s,e,m composes uniquement de caracteres '0' ou '1'
// Sortie : v = valeur correspondante
function v = valeur_reel64(s,e,m)
    // test des parametres en entree
    if type(s)~=10 then
        error("s doit etre une chaine de caracteres")
    end
    if prod(size(s))~=1 then
        error("s doit etre une chaine de caracteres unique")
    end
    if length(s)~=1 then
        error("s doit etre une chaine de 1 caractere")
    end
    if ~is_string_01(s) then
        error("s doit etre composee uniquement de ""0"" ou de ""1""")
    end
    
    if type(e)~=10 then
        error("e doit etre une chaine de caracteres")
    end
    if prod(size(e))~=1 then
        error("e doit etre une chaine de caracteres unique")
    end
    if length(e)~=11 then
        error("e doit etre une chaine de 11 caracteres")
    end
    if ~is_string_01(e) then
        error("e doit etre composee uniquement de ""0"" ou de ""1""")
    end
    
    
    if type(m)~=10 then
        error("m doit etre une chaine de caracteres")
    end
    if prod(size(m))~=1 then
        error("m doit etre une chaine de caracteres unique")
    end
    if length(m)~=52 then
        error("m doit etre une chaine de 52 caracteres")
    end
    if ~is_string_01(m) then
        error("m doit etre composee uniquement de ""0"" ou de ""1""")
    end
    
    
    // test valeur non reelle
    if e=="11111111111" then
        if m=="0000000000000000000000000000000000000000000000000000" then
            // valeur infinie
            if s=="0" then
                v = %inf;
            else
                v = -%inf;
            end
        else
            v = %nan
        end
        return;
    end
    
    // cas valeur réelle
    if e=="00000000000" then
        M = valeur_entier(m);
        // multiplier M par 2^(-1074)
        v = M/2^500;
        v = v/2^574;
    else
        M = valeur_entier("1"+m);
        E = valeur_entier(e)-1075;
        if E<-1000 then
            v = M*2^(-500);
            v = v*2^(E+500);
        else
            v = M*2^E; 
        end
    end
    
    // prise en compte du signe
    if s=="1" then
        v = -v;
    end
    
endfunction

///////////////////////////////////////////////////////////////////////////////
// écriture de la valeur exacte de la valeur v au format scientifique
function ecrire(v)
    // écriture de s sur 1100 caracteres
    s = msprintf("%1100.1093e", v)
    if length(s)<1100 then
        // ecrire s tel quel (s non reel)
        mprintf("%s\n", s)
    else
        // chercher la position ke du caractere "e"
        ke = 1100
        while part(s,ke)~="e"
            ke=ke-1
        end
        // supprimer les 0 inutiles à la fin de la mantisse
        k = ke-1;
        while part(s,k)=="0"
            k = k-1
        end
        s = part(s,1:k)+part(s,ke:1100);
        
        // ecrire s avec au plus 75 caracteres par ligne
        l_s = length(s);
        k1 = 1;
        while l_s-k1>75
            mprintf("%s\n",part(s,k1:k1+69))
            k1 = k1+70
        end
        mprintf("%s\n", part(s,k1:l_s))
    end
endfunction

//////////////////////////////////////////////////////////////////////////////
// ecriture du codage reel 64 chiffres binaires
// sous la forme signe (1 chiffre), exposant (11 c.), mantisse (52 c.)
function ecrire_codage64(v)
    // test des parametres en entree
    if type(v)~=1 then
        error("v doit etre une valeur numerique")
    end
    if length(v)~=1 then
        error("v doit etre un scalaire et non pas un tableau")
    end
    
    // codage
    [s,e,m] = codage_reel64(v);
    
    // ecriture
    mprintf("Codage 64 chiffres binaires :\n")
    mprintf(" signe s :\n  codage = %s\n  valeur = %d\n",s,valeur_entier(s))
    mprintf(" exposant e :\n  codage = %s\n  valeur = %d\n",e,valeur_entier(e))
    mprintf(" mantisse m :\n  codage = %s\n  valeur = %1.0f\n",m,valeur_entier(m))
endfunction
